﻿using System;
using System.Collections.Generic;
using System.Data.SqlClient;
using System.IO;
using System.Linq;
using System.Text;
using Microsoft.Build.Framework;
using Microsoft.Build.Utilities;
using Microsoft.SqlServer.Management.Common;
using Microsoft.SqlServer.Management.Smo;

namespace Sda.Build.Tasks
{
    //<UsingTask AssemblyFile="$(SolutionRoot)\BuildTools\Sda.Build.dll" TaskName="Sda.Build.Tasks.MssqlDBCreate" />
    //<MssqlDBCreate ServerName="fittest" DbName="Kreta" UserName="kreta" Password="kreta" MssqlScriptDir="$(SolutionRoot)\DBScripts\Mssql"/>
    public class MssqlDBCreate : Task
    {

        [Required]
        public string ServerName { get; set; }
        [Required]
        public string DbName { get; set; }

        [Required]
        public string UserName { get; set; }

        [Required]
        public string Password { get; set; }

        [Required]
        public string MssqlScriptDir { get; set; }

        public string DbConnection { get; set; }

        public string ManualDbName { get; set; }

        const string FunctionsDirectoryName = "dbo\\Functions";
        const string StoredProceduresDirectoryName = "dbo\\Stored procedures";
        const string DevStoredProceduresDirectoryName = "dev";
        const string MigrationDirectoryName = "..\\..\\Migrations";

        private Server dbServer;

        public Server DbServer
        {
            get
            {
                if (dbServer == null)
                {
                    if (string.IsNullOrEmpty(DbConnection))
                    {
                        var connectionStringBuilder = new SqlConnectionStringBuilder
                        {
                            IntegratedSecurity = false,
                            UserID = UserName,
                            Password = Password,
                            DataSource = ServerName
                        };

                        DbConnection = connectionStringBuilder.ToString();
                    }
                    var connection = new SqlConnection(DbConnection);

                    LogMessage("ConnectionString: " + DbConnection);

                    connection.Open();
                    dbServer = new Server(new ServerConnection(connection));
                }

                return dbServer;
            }
        }

        private string CreatedDbName { get; set; }

        public override bool Execute()
        {
            try
            {
                // ha létezik, régi adatbázis eldobása és új létrehozása
                bool isNewDb = CreateDatabase();

                // ha létezik a DB (vagyis nem változott a verziószám), akkor nem fog új DB-t generálni.
                if (isNewDb)
                {
                    // adatbázis séma kialakítása (táblák, indexek, constraintek, tárolt eljárások...)
                    CreatingDatabaseSchema();

                }
            }
            catch (Exception ex)
            {
                if (BuildEngine != null)
                {
                    Log.LogErrorFromException(ex);
                }
            }

            return !Log.HasLoggedErrors;
        }

        /// <summary>
        /// Csak akkor dobja el és hozza újra létre a Db-t, ha változott a verziószám.
        /// </summary>
        /// <returns></returns>
        private bool CreateDatabase()
        {
            if (!string.IsNullOrEmpty(DbConnection))
            {
                CreatedDbName = DbName;

                if (DbServer.Databases[CreatedDbName] != null)
                {
                    LogMessage($"{CreatedDbName} database exists!");
                    return false;
                }
            }
            else if (string.IsNullOrEmpty(ManualDbName))
            {
                CreatedDbName = DbName;

                if (DbServer.Databases[CreatedDbName] != null)
                {
                    LogMessage($"{CreatedDbName} database exists!");
                    return false;
                }
            }
            else
            {
                CreatedDbName = $"{DbName}_{ManualDbName}";

                if (DbServer.Databases[CreatedDbName] != null)
                {
                    LogMessage($"Dropping existing {CreatedDbName} on {ServerName} ...");
                    DbServer.KillAllProcesses(CreatedDbName);

                    DbServer.Databases[CreatedDbName].Drop();
                }
            }

            LogMessage($"Creating {CreatedDbName} on {ServerName} ...");

            var newDb = new Database(DbServer, CreatedDbName)
            {
                Collation = "Hungarian_CI_AI"
            };
            newDb.Create();

            LogMessage("Done.");
            return true;
        }

        private void CreatingDatabaseSchema()
        {
            var scriptList = new List<string>();

            scriptList.AddRange(Directory.GetFiles(MssqlScriptDir, "create.schemas.sql", SearchOption.TopDirectoryOnly).ToList());
            scriptList.AddRange(Directory.GetFiles(MssqlScriptDir, "create.schemas.*.sql", SearchOption.TopDirectoryOnly).ToList());   // schémák létrehozása

            scriptList.AddRange(Directory.GetFiles(MssqlScriptDir, "create.tables.sql", SearchOption.TopDirectoryOnly).ToList());
            scriptList.AddRange(Directory.GetFiles(MssqlScriptDir, "create.tables.*.sql", SearchOption.TopDirectoryOnly).ToList());    // táblák létrehozása

            scriptList.AddRange(Directory.GetFiles(MssqlScriptDir, "alter.tables.sql", SearchOption.TopDirectoryOnly).ToList());
            scriptList.AddRange(Directory.GetFiles(MssqlScriptDir, "alter.tables.*.sql", SearchOption.TopDirectoryOnly).ToList());     // táblák módosítása

            scriptList.AddRange(Directory.GetFiles(MssqlScriptDir, "create.comments.sql", SearchOption.TopDirectoryOnly).ToList());
            scriptList.AddRange(Directory.GetFiles(MssqlScriptDir, "create.comments.*.sql", SearchOption.TopDirectoryOnly).ToList());  // táblák, oszlopok descriptionjei

            scriptList.AddRange(Directory.GetFiles(MssqlScriptDir, "create.indexes.sql", SearchOption.TopDirectoryOnly).ToList());
            scriptList.AddRange(Directory.GetFiles(MssqlScriptDir, "create.indexes.*.sql", SearchOption.TopDirectoryOnly).ToList());  // indexek létrehozása

            scriptList.AddRange(Directory.GetFiles(MssqlScriptDir, "create.triggers.sql", SearchOption.TopDirectoryOnly).ToList());
            scriptList.AddRange(Directory.GetFiles(MssqlScriptDir, "create.triggers.*.sql", SearchOption.TopDirectoryOnly).ToList());  // triggerek létrehozása

            scriptList.AddRange(Directory.GetFiles(MssqlScriptDir, "create.data.sql", SearchOption.TopDirectoryOnly).ToList());
            scriptList.AddRange(Directory.GetFiles(MssqlScriptDir, "create.data.*.sql", SearchOption.TopDirectoryOnly).ToList());      // alapadatok beszúrása (insert)

            scriptList.AddRange(Directory.GetFiles(MssqlScriptDir, "create.testdata.sql", SearchOption.TopDirectoryOnly).ToList());
            scriptList.AddRange(Directory.GetFiles(MssqlScriptDir, "create.testdata.*.sql", SearchOption.TopDirectoryOnly).ToList());  // alap tesztadatok beszúrása (insert)


            ExecureSqlScripts(scriptList);

            CreateDevStoredProcedures();
            CreateGlobalStoredProcedures();

            scriptList = new List<string>();

            scriptList.AddRange(Directory.GetFiles(MssqlScriptDir, "create.views.sql", SearchOption.TopDirectoryOnly).ToList());
            scriptList.AddRange(Directory.GetFiles(MssqlScriptDir, "create.views.*.sql", SearchOption.TopDirectoryOnly).ToList());     // dbo-s view-k létrehozása

            scriptList.AddRange(Directory.GetFiles(MssqlScriptDir, "create.auditlog.sql", SearchOption.TopDirectoryOnly).ToList());
            scriptList.AddRange(Directory.GetFiles(MssqlScriptDir, "create.auditlog.*.sql", SearchOption.TopDirectoryOnly).ToList());   // auditlog létrehozása

            ExecureSqlScripts(scriptList);

            CreateFunctions();
            CreateStoredProcedures();
            VersionInfoInsert();


        }

        private void VersionInfoInsert()
        {
            string script = "";
            LogMessage($"Populate VersionInfo");

            var dirs = Directory.GetFiles(Path.Combine(MssqlScriptDir, MigrationDirectoryName, "Archive"), "20*.cs", SearchOption.TopDirectoryOnly).Select(d => Path.GetFileNameWithoutExtension(d)).ToList();
            dirs.AddRange(Directory.GetFiles(Path.Combine(MssqlScriptDir, MigrationDirectoryName), "20*.cs", SearchOption.TopDirectoryOnly).Select(d => Path.GetFileNameWithoutExtension(d)).ToList());

            int i = 0;

            foreach (var dir in dirs)
            {

                if (i++ == 0)
                {
                    script = "INSERT INTO VersionInfo (Version, AppliedOn, Description) VALUES\n";
                    script += "   (" + dir.Substring(0, 14) + ", GETDATE(), '" + dir.Substring(15) + "')\n";
                }
                else
                {
                    script += "  ,(" + dir.Substring(0, 14) + ", GETDATE(), '" + dir.Substring(15) + "')\n";
                }
                if (i == 500)
                {
                    i = 0;
                    try
                    {
                        Database db = DbServer.Databases[CreatedDbName];
                        db.ExecuteNonQuery(script);

                    }
                    catch (Exception ex)
                    {
                        LogMessage($"Hiba az SQL script futtatása közben: Populate VersionInfo, hibaüzenet: {ex.Message}");
                        throw new Exception($"Hiba az SQL script futtatása közben: Populate VersionInfo, hibaüzenet: {ex.Message}");
                    }

                }
            }
        }

        private void ExecureSqlScripts(List<String> scriptList)
        {
            var hasError = false;

            foreach (var scriptName in scriptList)
            {
                try
                {
                    var flieName = Path.GetFileName(scriptName);
                    LogMessage($"Executing script {flieName} ...");
                    ExecuteScript(scriptName);

                }
                catch (Exception ex)
                {
                    LogMessage($"Hiba az SQL script futtatása közben: {scriptName}, hibaüzenet: {ex.Message}");
                    throw new Exception($"Hiba az SQL script futtatása közben: {scriptName}, hibaüzenet: {ex.Message}");
                }
            }

            if (hasError)
            {
                throw new Exception("Hibás SQL script!");
            }
        }

        private void CreateGlobalStoredProcedures()
        {
            LogMessage("Creating global stored procedures ...");
            var hasError = false;

            var storedProcedureFiles = Directory.GetFiles(MssqlScriptDir, "sp_Global_*.sql", SearchOption.TopDirectoryOnly);
            foreach (var storedProcedureFile in storedProcedureFiles)
            {
                try
                {
                    ExecuteScript(storedProcedureFile);
                }
                catch (Exception ex)
                {
                    LogMessage($"Hibás tárolt eljárás: {storedProcedureFile}, hibaüzenet: {ex.Message}");
                    hasError = true;
                }
            }

            if (hasError)
            {
                throw new Exception("Hibás tárolt eljárások!");
            }
            LogMessage("Done.");
        }

        private void CreateStoredProcedures()
        {
            LogMessage("Creating stored procedures ...");
            var hasError = false;

            var storedProcedureFiles = Directory.GetFiles(Path.Combine(MssqlScriptDir, StoredProceduresDirectoryName), "*.sql", SearchOption.AllDirectories);
            foreach (var storedProcedureFile in storedProcedureFiles)
            {
                try
                {
                    ExecuteScript(storedProcedureFile);
                }
                catch (Exception ex)
                {
                    LogMessage($"Hibás tárolt eljárás: {storedProcedureFile}, hibaüzenet: {ex.Message}");
                    hasError = true;
                }
            }

            if (hasError)
            {
                throw new Exception("Hibás tárolt eljárások!");
            }
            LogMessage("Done.");
        }

        private void CreateDevStoredProcedures()
        {
            LogMessage("Creating dev schema stored procedures and functions ...");
            var hasError = false;

            var storedProcedureFiles = Directory.GetFiles(Path.Combine(MssqlScriptDir, DevStoredProceduresDirectoryName), "*.sql", SearchOption.AllDirectories);
            foreach (var storedProcedureFile in storedProcedureFiles)
            {
                try
                {
                    ExecuteScript(storedProcedureFile);
                }
                catch (Exception ex)
                {
                    LogMessage($"Hibás tárolt eljárás: {storedProcedureFile}, hibaüzenet: {ex.Message}");
                    hasError = true;
                }
            }

            if (hasError)
            {
                throw new Exception("Hibás tárolt eljárások!");
            }
            LogMessage("Done.");
        }


        private void CreateFunctions()
        {
            LogMessage("Creating functions ...");
            var hasError = false;


            string functionPath;
            string line;

            functionPath = Path.Combine(MssqlScriptDir, FunctionsDirectoryName);

            using (var reader = new StreamReader(Path.Combine(functionPath, "FunctionOrder.txt"), Encoding.UTF8))
            {
                while ((line = reader.ReadLine()) != null)
                {
                    try
                    {
                        LogMessage($"Creating sorted functions ... {line}");
                        ExecuteScript(Path.Combine(functionPath, line));
                    }
                    catch (Exception ex)
                    {
                        LogMessage($"Hibás függvény: {line}, hibaüzenet: {ex.Message}");
                        hasError = true;
                    }
                }
            }

            var functionFiles = Directory.GetFiles(functionPath, "*.sql", SearchOption.AllDirectories);
            foreach (var functionFile in functionFiles)
            {
                try
                {
                    ExecuteScript(functionFile);
                }
                catch (Exception ex)
                {
                    LogMessage($"Hibás függvény: {functionFile}, hibaüzenet: {ex.Message}");
                    hasError = true;
                }
            }

            if (hasError)
            {
                throw new Exception("Hibás függvények!");
            }
            LogMessage("Done.");
        }


        private void ExecuteScript(string scriptPath)
        {
            string script;
            using (var reader = new StreamReader(scriptPath, Encoding.UTF8))
            {
                script = reader.ReadToEnd();
            }

            Database db = DbServer.Databases[CreatedDbName];
            db.ExecuteNonQuery(script);
        }

        private void LogMessage(string message)
        {
            LogMessage(MessageImportance.High, message);
        }

        private void LogMessage(MessageImportance importance, string message)
        {
            if (BuildEngine != null)
            {
                Log.LogMessage(importance, message);
            }
        }

    }
}
