﻿using System;
using System.Diagnostics;
using System.Diagnostics.CodeAnalysis;
using System.Diagnostics.Contracts;
using JetBrains.Annotations;
using Microsoft.Build.Framework;
using Microsoft.Build.Utilities;

namespace Sda.Build.Tasks
{
    static class TaskExtensions
    {
        static readonly TraceSource _trace = new TraceSource("TaskLog", SourceLevels.Off);

        static bool CheckLog([NotNull] Task task)
        {
            // ReSharper disable once InvocationIsSkipped
            Contract.Requires(task != null);

            return
                task.HostObject != null
                ||
                task.BuildEngine != null
                ||
                task.BuildEngine2 != null
                ||
                task.BuildEngine3 != null
                ||
                task.BuildEngine4 != null;
        }

        [StringFormatMethod("message")]
        public static void LogHigh([NotNull] this Task task, [NotNull] string message, [CanBeNull] params object[] args)
        {
            if (task == null)
                throw new ArgumentNullException("task");
            if (message == null)
                throw new ArgumentNullException("message");

            if (CheckLog(task))
            {
                task.Log.LogMessage(MessageImportance.High, message.Replace("\r", ""), args);
            }
            else
            {
                _trace.TraceEvent(TraceEventType.Warning, 0, message, args ?? new object[0]);
            }
        }

        [StringFormatMethod("message")]
        public static void LogInfo([NotNull] this Task task, [NotNull] string message, [CanBeNull] params object[] args)
        {
            if (task == null)
                throw new ArgumentNullException("task");
            if (message == null)
                throw new ArgumentNullException("message");

            if (CheckLog(task))
            {
                task.Log.LogMessage(MessageImportance.Normal, message.Replace("\r", ""), args);
            }
            else
            {
                _trace.TraceEvent(TraceEventType.Information, 0, message, args ?? new object[0]);
            }
        }

        [StringFormatMethod("message")]
        public static void LogWarning([NotNull] this Task task, [NotNull] string message, [CanBeNull] params object[] args)
        {
            if (task == null)
                throw new ArgumentNullException("task");
            if (message == null)
                throw new ArgumentNullException("message");

            if (CheckLog(task))
            {
                task.Log.LogWarning(message.Replace("\r", ""), args);
            }
            else
            {
                _trace.TraceEvent(TraceEventType.Warning, 0, message, args ?? new object[0]);
            }
        }

        [StringFormatMethod("message")]
        [SuppressMessage("Microsoft.Performance", "CA1811:AvoidUncalledPrivateCode")]
        public static void LogError([NotNull] this Task task, [NotNull] string message, [CanBeNull] params object[] args)
        {
            if (task == null)
                throw new ArgumentNullException("task");
            if (message == null)
                throw new ArgumentNullException("message");

            if (CheckLog(task))
            {
                task.Log.LogError(message.Replace("\r", ""), args);
            }
            else
            {
                _trace.TraceEvent(TraceEventType.Error, 0, message, args ?? new object[0]);
            }
        }

        [SuppressMessage("Microsoft.Performance", "CA1811:AvoidUncalledPrivateCode")]
        public static void LogError([NotNull] this Task task, [NotNull] Exception exception)
        {
            if (task == null)
                throw new ArgumentNullException("task");
            if (exception == null)
                throw new ArgumentNullException("exception");

            LogError(task, exception.ToString());
        }
    }
}
