﻿using System;
using System.Collections.Generic;
using System.Diagnostics;
using System.Diagnostics.CodeAnalysis;
using System.Globalization;
using System.IO;
using System.Linq;
using System.Runtime.CompilerServices;
using System.Text;
using System.Xml;
using System.Xml.Linq;
using JetBrains.Annotations;

namespace Sda.ModelIdUpdater
{
    public sealed class Updater
    {
        static readonly XNamespace _uml13 = "omg.org/UML1.3";

        static readonly HashSet<string> _classTags = new HashSet<string>
        {
            "allread", "SavedDI", "historymode", "disabledsecuritymode", "anonymous", "PoszeidonDeleter", "dkt", "inheritance", "dktnameset"
        };

        static readonly HashSet<string> _attributeTags = new HashSet<string>
        {
            "required", "length", "precision", "readonly", "trim", "defaultonly", "multilang", "LanguageIndex",
            "history", "minvalue", "maxvalue", "minlength", "maxlength", "truncateable", "value", "index_function",
            "Unique", "Unique1", "Unique2", "index", "index1", "index2", "unicode", "dictdefault", "computed", "persisted"
        };

        static readonly HashSet<string> _operationTags = new HashSet<string>
        {
            "progress"
        };

        static readonly HashSet<string> _associationTags = new HashSet<string>
        {
            "defaultonly", "cascadedelete", "readonly"
        };

        static readonly HashSet<string> _classStereotypes = new HashSet<string>
        {
            "SDAEntity"
        };

        static readonly HashSet<string> _attributeStereotypes = new HashSet<string>
        {
            "DisplayName"
        };

        static readonly HashSet<string> _operationStereotypes = new HashSet<string>
        {
            "Common", "Neptun", "Poszeidon"
        };


        static readonly HashSet<string> _multiplicities = new HashSet<string>
        {
            "0..1", "0..*", "*", "1", "1..*"
        };

        static void UpdateAssociationEnd(XContainer associationEnd, ref long lastId)
        {
            if (associationEnd == null)
            {
                throw new ArgumentNullException("associationEnd");
            }
            var tagged =
                associationEnd
                .Element(_uml13 + "ModelElement.taggedValue");
            if (tagged == null)
            {
                tagged = new XElement("ModelElement.taggedValue");
                associationEnd.Add(tagged);
            }
            var style =
                tagged
                .Elements(_uml13 + "TaggedValue")
                .Where(e => e.Attributes().Any(a => a.Name == "tag" && a.Value == @"deststyle"))
                .Select(e => e.Attributes().Single(a => a.Name == "value"))
                .FirstOrDefault();
            // ReSharper disable once ConvertIfStatementToNullCoalescingExpression
            if (style == null)
            {
                style =
                    tagged
                    .Elements(_uml13 + "TaggedValue")
                    .Where(e => e.Attributes().Any(a => a.Name == "tag" && a.Value == "sourcestyle"))
                    .Select(e => e.Attributes().Single(a => a.Name == "value"))
                    .FirstOrDefault();
            }
            if (style == null)
            {
                style = new XAttribute("value", "");
                tagged.Add(
                    new XElement(
                        _uml13 + "TaggedValue",
                         new XAttribute("tag", "sourcestyle"),
                        style));
            }
            // ReSharper disable once InvertIf
            if (style.Value.IndexOf("alias=", StringComparison.OrdinalIgnoreCase) == -1)
            {
                lastId++;
                style.Value += string.Format(CultureInfo.InvariantCulture, "alias={0};", lastId);
            }
        }

        static void UpdateAssociationName(XElement association)
        {
            if (association == null)
                throw new ArgumentNullException("association");

            var nameAttribute = association.Attribute("name");
            if (nameAttribute != null)
            {
                return;
            }
            string left =
                association
                .Descendants(_uml13 + "TaggedValue")
                .Where(e => e.Attribute("tag") != null && e.Attribute("value") != null && e.Attribute("tag").Value == "ea_sourceName")
                .Select(e => e.Attribute("value").Value)
                .FirstOrDefault();
            string right =
                association
                .Descendants(_uml13 + "TaggedValue")
                .Where(e => e.Attribute("tag") != null && e.Attribute("value") != null && e.Attribute("tag").Value == "ea_targetName")
                .Select(e => e.Attribute("value").Value)
                .FirstOrDefault();
            var ends = association.Descendants(_uml13 + "AssociationEnd").Take(2).ToList();
            if (ends.Count == 2)
            {
                left = ends[0].Attribute("name") == null ? left : ends[0].Attribute("name").Value;
                right = ends[1].Attribute("name") == null ? right : ends[1].Attribute("name").Value;
            }
            if (string.IsNullOrEmpty(left)
                ||
                string.IsNullOrEmpty(right))
            {
                return;
            }
            nameAttribute = new XAttribute("name", string.Format(CultureInfo.InvariantCulture, "{0}_{1}", left, right));
            association.Add(nameAttribute);
        }

        static void UpdateFeature(XContainer feature, ref long lastId)
        {
            var tagged = feature.Element(_uml13 + "ModelElement.taggedValue");
            if (tagged == null)
            {
                tagged = new XElement("ModelElement.taggedValue");
                feature.Add(tagged);
            }
            var alias =
                tagged
                .Elements(_uml13 + "TaggedValue")
                .Where(e => e.Attributes().Any(a => a.Name == "tag" && a.Value == "style"))
                .Select(e => e.Attributes().Single(a => a.Name == "value"))
                .FirstOrDefault();
            if (alias == null)
            {
                alias = new XAttribute("value", "");
                tagged.Add(
                    new XElement(
                        _uml13 + "TaggedValue",
                        new XAttribute("tag", "style"),
                        alias));
            }
            // ReSharper disable once InvertIf
            if (string.IsNullOrWhiteSpace(alias.Value))
            {
                lastId++;
                alias.Value = string.Format(CultureInfo.InvariantCulture, "{0}", lastId);
            }
        }

        static void UpdateGeneralization(XContainer generalization, ref long lastId)
        {
            if (generalization == null)
            {
                throw new ArgumentNullException("generalization");
            }
            var tagged =
                generalization
                .Element(_uml13 + "ModelElement.taggedValue");
            if (tagged == null)
            {
                tagged = new XElement("ModelElement.taggedValue");
                generalization.Add(tagged);
            }
            var styleex =
                tagged
                .Elements(_uml13 + "TaggedValue")
                .Where(e => e.Attributes().Any(a => a.Name == "tag" && a.Value == "styleex"))
                .Select(e => e.Attributes().Single(a => a.Name == "value"))
                .FirstOrDefault();
            if (styleex == null)
            {
                styleex = new XAttribute("value", "");
                tagged.Add(
                    new XElement(
                        _uml13 + "TaggedValue",
                        new XAttribute("tag", "styleex"),
                        styleex));
            }
            if (styleex.Value.IndexOf("alias=", StringComparison.OrdinalIgnoreCase) != -1)
            {
                return;
            }
            lastId++;
            styleex.Value += string.Format(CultureInfo.InvariantCulture, "alias={0};", lastId);
        }

        static IEnumerable<string> FindDictionaryItems(XContainer content)
        {
            if (content == null)
                throw new ArgumentNullException("content");

            return from element in content.Descendants(_uml13 + "Generalization")
                   let source = element
                        .Descendants(_uml13 + "TaggedValue")
                        .Where(t => t.Attribute("tag") != null && t.Attribute("tag").Value == "ea_sourceName")
                        .Select(t => t.Attribute("value").Value)
                        .FirstOrDefault()
                   let destination = element
                        .Descendants(_uml13 + "TaggedValue")
                        .Where(t => t.Attribute("tag") != null && t.Attribute("tag").Value == "ea_targetName")
                        .Select(t => t.Attribute("value").Value)
                        .FirstOrDefault()
                   where !string.IsNullOrEmpty(source)
                         &&
                         !string.IsNullOrEmpty(destination)
                         &&
                        (destination.Equals("DictionaryItemBase", StringComparison.OrdinalIgnoreCase) ||
                        destination.Equals("DKT_DictionaryItemBase", StringComparison.OrdinalIgnoreCase))
                   select source;
        }

        static string GetFullNameAssoc(XContainer content, XElement element, out string package)
        {
            if (element == null
                ||
                element.Name != _uml13 + "Association"
                ||
                element.Attribute("name") == null)
            {
                package = null;
                return null;
            }
            string name = element.Attribute("name").Value;
            string source =
                element
                .Descendants(_uml13 + "TaggedValue")
                .Where(e =>
                   e.Attribute("tag") != null
                   &&
                   e.Attribute("tag").Value == "ea_sourceName"
                   &&
                   e.Attribute("value") != null)
                .Select(e => e.Attribute("value").Value)
                .FirstOrDefault();
            string target =
                element
                .Descendants(_uml13 + "TaggedValue")
                .Where(e =>
                   e.Attribute("tag") != null
                   &&
                   e.Attribute("tag").Value == "ea_targetName"
                   &&
                   e.Attribute("value") != null)
                .Select(e => e.Attribute("value").Value)
                .FirstOrDefault();
            if (!string.IsNullOrEmpty(source)
                &&
                !string.IsNullOrEmpty(target))
            {
                name += " (" + source + "->" + target + ")";
            }
            string sourcePackage =
                content
                .Descendants(_uml13 + "Class")
                .Where(e =>
                   e.Attribute("name") != null
                   &&
                   e.Attribute("namespace") != null
                   &&
                   e.Attribute("name").Value == source)
                .Select(e => e.Attribute("namespace").Value.Replace("EAPK_", ""))
                .FirstOrDefault();
            string targetPackage =
                content
                .Descendants(_uml13 + "Class")
                .Where(e =>
                   e.Attribute("name") != null
                   &&
                   e.Attribute("namespace") != null
                   &&
                   e.Attribute("name").Value == target)
                .Select(e => e.Attribute("namespace").Value.Replace("EAPK_", ""))
                .FirstOrDefault();
            if (sourcePackage == targetPackage)
            {
                package = sourcePackage;
                name = package + "::" + name;
            }
            else if (sourcePackage == "Common")
            {
                package = targetPackage;
                name = package + "::" + name;
            }
            else if (targetPackage == "Common")
            {
                package = sourcePackage;
                name = package + "::" + name;
            }
            else
            {
                package = null;
            }
            return name;
        }

        static string GetFullName(XElement element, out string package)
        {
            if (element == null
                ||
                element.Attribute("name") == null)
            {
                package = null;
                return null;
            }
            string name = element.Attribute("name").Value;
            XElement @class;
            if (element.Parent != null
                &&
                (@class = element.Parent.Parent) != null
                &&
                @class.Name == _uml13 + "Class"
                &&
                @class.Attribute("name") != null)
            {
                name = @class.Attribute("name").Value + "." + name;
            }
            else
            {
                @class = element;
            }
            XElement @namespace;
            if ((@namespace = element.Descendants(_uml13 + "Stereotype").FirstOrDefault()) != null
                &&
                _operationStereotypes.Contains((string)@namespace.Attribute("name")))
            {
                package = (string)@namespace.Attribute("name");
                name = package + "::" + name;
                return name;
            }
            if (@class.Parent != null
                &&
                (@namespace = @class.Parent.Parent) != null
                &&
                @namespace.Name == _uml13 + "Package"
                &&
                @namespace.Attribute("name") != null)
            {
                package = (string)@namespace.Attribute("name");
                name = package + "::" + name;
                return name;
            }
            package = null;
            return name;
        }

        readonly Dictionary<long, string> _ids = new Dictionary<long, string>();

        public event EventHandler<LogEventArgs> LogMessageHandler;

        void CheckId(XElement tag)
        {
            if (tag == null)
                throw new ArgumentNullException("tag");

            string value = (string)tag.Attribute("value");
            long id = 0;
            if (string.IsNullOrWhiteSpace(value))
            {
                return;
            }
            bool found = false;
            // ReSharper disable once LoopCanBeConvertedToQuery
            foreach (string members in
                value
                .Split(new[] { ';' }, StringSplitOptions.RemoveEmptyEntries)
                .Where(m => m.StartsWith("alias=", StringComparison.OrdinalIgnoreCase)))
            {
                if (!long.TryParse(members.Substring(6), out id)
                    ||
                    id < 10)
                {
                    continue;
                }
                found = true;
                break;
            }
            if (!found
                &&
                (
                    !long.TryParse(value, out id)
                    ||
                    id < 10
                ))
            {
                return;
            }
            string newOwner = "";
            XElement owner;
            if (tag.Parent != null
                &&
                (owner = tag.Parent.Parent) != null
                &&
                owner.Attribute("name") != null)
            {
                newOwner = owner.Attribute("name").Value;
            }
            string oldOwner;
            if (_ids.TryGetValue(id, out oldOwner))
            {
                if (LogMessageHandler != null)
                {
                    LogMessageHandler(
                        null,
                        new LogEventArgs
                        {
                            EventType = TraceEventType.Error,
                            Package = "Common",
                            Message = string.Format(
                                CultureInfo.InvariantCulture,
                                "Duplication error, same id ({0}) found for both {1} and {2}!",
                                id,
                                oldOwner,
                                newOwner)
                        });
                }
            }
            else
            {
                _ids.Add(id, newOwner);
            }
        }

        void FindLastIdAlias(XContainer content)
        {
            if (content == null)
                throw new ArgumentNullException("content");

            foreach (var tag in
                content
                .Descendants(_uml13 + "TaggedValue")
                .Where(tag => (string)tag.Attribute("tag") == "alias"))
            {
                CheckId(tag);
            }
        }

        void FindLastIdStyle(XContainer content)
        {
            if (content == null)
                throw new ArgumentNullException("content");

            foreach (var tag in
                content
                .Descendants(_uml13 + "TaggedValue")
                .Where(tag => (string)tag.Attribute("tag") == "style"))
            {
                CheckId(tag);
            }
        }

        void FindLastIdStyleEx(XContainer content)
        {
            if (content == null)
                throw new ArgumentNullException("content");

            // ReSharper disable once LoopCanBePartlyConvertedToQuery
            foreach (var tag in
                content
                .Descendants(_uml13 + "TaggedValue")
                .Where(tag =>
                   (string)tag.Attribute("tag") == "styleex"
                   ||
                   (string)tag.Attribute("tag") == "sourcestyle"
                   ||
                   (string)tag.Attribute("tag") == @"deststyle"))
            {
                CheckId(tag);
            }
        }

        long FindLastId(XContainer content)
        {
            if (content == null)
                throw new ArgumentNullException("content");

            FindLastIdAlias(content);
            FindLastIdStyle(content);
            FindLastIdStyleEx(content);
            return _ids.Keys.Max();
        }


        [SuppressMessage("Microsoft.Globalization", "CA1303:Do not pass literals as localized parameters")]
        void VerifyDictionaries(XContainer content, ICollection<string> dictionaries)
        {
            if (content == null)
                throw new ArgumentNullException("content");
            if (dictionaries == null)
                throw new ArgumentNullException("dictionaries");

            if (LogMessageHandler == null)
            {
                return;
            }
            var dictionaryAttributes =
                content
                .Descendants(_uml13 + "Attribute")
                .Where(at =>
                   at
                   .Descendants(_uml13 + "TaggedValue")
                   .Any(tv =>
                      tv.Attribute("tag") != null
                      &&
                      tv.Attribute("value") != null
                      &&
                      tv.Attribute("tag").Value.Equals("type", StringComparison.OrdinalIgnoreCase)
                      &&
                      tv.Attribute("value").Value.Equals("DictionaryItem", StringComparison.OrdinalIgnoreCase)));
            foreach (var attribute in dictionaryAttributes)
            {
                string package, name = GetFullName(attribute, out package);
                var initValue =
                    attribute.Descendants(_uml13 + "Attribute.initialValue")
                    .Elements(_uml13 + "Expression")
                    .Attributes("body")
                    .FirstOrDefault();
                if (initValue == null)
                {
                    LogMessageHandler(
                        null,
                        new LogEventArgs
                        {
                            EventType = TraceEventType.Error,
                            Package = package,
                            Message = string.Format(
                                CultureInfo.InvariantCulture,
                                "Dictionary attribute '{0}' is missing dictionary type!",
                                name)
                        });
                }
                else if (!dictionaries.Contains(initValue.Value))
                {
                    LogMessageHandler(
                        null,
                        new LogEventArgs
                        {
                            EventType = TraceEventType.Error,
                            Package = package,
                            Message = string.Format(
                                CultureInfo.InvariantCulture,
                                "Dictionary attribute '{0}' has invalid dictionary type '{1}'!",
                                name,
                                initValue.Value)
                        });
                }
            }
        }

        [SuppressMessage("Microsoft.Globalization", "CA1303:Do not pass literals as localized parameters")]
        [UsedImplicitly]
        void VerifyUniquePairs(XContainer content)
        {
            if (content == null)
                throw new ArgumentNullException("content");

            if (LogMessageHandler == null)
            {
                return;
            }
            // ReSharper disable once LoopCanBePartlyConvertedToQuery
            foreach (var association in content.Descendants(_uml13 + "Association"))
            {
                int twoEnds =
                    association
                    .Descendants(_uml13 + "AssociationEnd")
                    .Count(a =>
                       a.Attribute("multiplicity") != null
                       &&
                       a.Attribute("multiplicity").Value == "0..1");
                if (twoEnds < 2)
                {
                    continue;
                }
                string package, name = GetFullNameAssoc(content, association, out package);
                if (string.IsNullOrEmpty(name))
                {
                    continue;
                }
                LogMessageHandler(
                    null,
                    new LogEventArgs
                    {
                        EventType = TraceEventType.Warning,
                        Package = package,
                        Message = string.Format(
                            CultureInfo.InvariantCulture,
                            "Association '{0}' [0..1-0..1] should be changed to [0..1-*]!",
                            name)
                    });
            }
        }

        [SuppressMessage("Microsoft.Globalization", "CA1303:Do not pass literals as localized parameters")]
        void VerifyMultiplicity(XContainer content)
        {
            if (content == null)
                throw new ArgumentNullException("content");

            if (LogMessageHandler == null)
            {
                return;
            }
            var associationEnds =
                content
                .Descendants(_uml13 + "AssociationEnd")
                .Where(a =>
                   a.Attribute("multiplicity") == null
                   ||
                   !_multiplicities.Contains(a.Attribute("multiplicity").Value))
                .Where(a =>
                   a.Parent != null
                   &&
                   a.Parent.Parent != null);
            foreach (var associationEnd in associationEnds)
            {
                string multiplicity = "?";
                if (associationEnd.Attribute("multiplicity") != null)
                {
                    multiplicity = associationEnd.Attribute("multiplicity").Value;
                }
                string package = null, name = null;
                if (associationEnd.Parent != null
                    &&
                    associationEnd.Parent.Parent != null)
                {
                    name = GetFullName(associationEnd.Parent.Parent, out package);
                }
                if (string.IsNullOrEmpty(name))
                {
                    return;
                }
                LogMessageHandler(
                    null,
                    new LogEventArgs
                    {
                        EventType = TraceEventType.Error,
                        Package = package,
                        Message = string.Format(
                            CultureInfo.InvariantCulture,
                            "Association '{0}' has invalid multiplicity '{1}'!",
                            name,
                            multiplicity)
                    });
            }
        }

        [SuppressMessage("Microsoft.Globalization", "CA1303:Do not pass literals as localized parameters")]
        void VerifyNotes(XContainer content, XElement element, string type)
        {
            if (element == null)
                throw new ArgumentNullException("element");
            if (string.IsNullOrEmpty(type))
                throw new ArgumentNullException("type");

            if (LogMessageHandler == null)
            {
                return;
            }
            string package, name = type == "Association"
                ? GetFullNameAssoc(content, element, out package)
                : GetFullName(element, out package);
            if (string.IsNullOrEmpty(name))
            {
                return;
            }
            if (string.IsNullOrEmpty(package))
            {
                LogMessageHandler(
                    null,
                    new LogEventArgs
                    {
                        EventType = TraceEventType.Error,
                        Package = "Common",
                        Message = string.Format(
                            CultureInfo.InvariantCulture,
                            "Couldn't resolve package owner of {0}!",
                            name)
                    });
            }
            /*if( element
                .Elements( _uml13 + "ModelElement.taggedValue" )
                .Descendants( _uml13 + "TaggedValue" )
                .Any( tv =>
                        tv.Attribute( "tag" ) != null
                        &&
                        ( tv.Attribute( "tag" ).Value == "documentation" || tv.Attribute( "tag" ).Value == "description" )
                        &&
                        tv.Attribute( "value" ) != null
                        &&
                        !string.IsNullOrWhiteSpace( tv.Attribute( "value" ).Value ) ) )
            {
                return;
            }
            LogMessageHandler(
                null,
                new LogEventArgs
                {
                    EventType = TraceEventType.Warning,
                    Package = package,
                    Message = string.Format(
                        CultureInfo.InvariantCulture,
                        "{0} '{1}' has no documentation!",
                        type,
                        name )
                } );*/
        }

        [SuppressMessage("Microsoft.Globalization", "CA1303:Do not pass literals as localized parameters")]
        [SuppressMessage("ReSharper", "SuggestBaseTypeForParameter")]
        void VerifyTag(XElement element, string type, HashSet<string> validTags)
        {
            if (element == null)
                throw new ArgumentNullException("element");
            if (string.IsNullOrEmpty(type))
                throw new ArgumentNullException("type");
            if (validTags == null)
                throw new ArgumentNullException("validTags");

            if (LogMessageHandler == null)
            {
                return;
            }
            string package, name = GetFullName(element, out package);
            if (string.IsNullOrEmpty(name))
            {
                return;
            }
            var invalidGlobalTags = Enumerable.Empty<string>();
            XDocument doc;
            if (element.Attribute("xmi.id") != null
                &&
                (doc = element.Document) != null
                &&
                doc.Root != null)
            {
                var elementId = element.Attribute("xmi.id").Value;
                invalidGlobalTags =
                    doc.Root
                    .Elements("XMI.content")
                    .Elements(_uml13 + "TaggedValue")
                    .Where(tv =>
                       tv.Attribute("tag") != null
                       &&
                       tv.Attribute("value") != null
                       &&
                       tv.Attribute("xmi.id") != null
                       &&
                       tv.Attribute("modelElement") != null
                       &&
                       tv.Attribute("modelElement").Value == elementId
                       &&
                       !validTags.Contains(tv.Attribute("tag").Value))
                     .Select(tv => tv.Attribute("tag").Value);
            }
            var invalidTags =
                element
                .Elements(_uml13 + "ModelElement.taggedValue")
                .Descendants(_uml13 + "TaggedValue")
                .Where(tv =>
                   tv.Attribute("tag") != null
                   &&
                   tv.Attribute("value") != null
                   &&
                   tv.Attribute("xmi.id") != null
                   &&
                   !validTags.Contains(tv.Attribute("tag").Value))
                 .Select(tv => tv.Attribute("tag").Value);
            foreach (var tag in invalidTags.Concat(invalidGlobalTags))
            {
                LogMessageHandler(
                    null,
                    new LogEventArgs
                    {
                        EventType = TraceEventType.Error,
                        Package = package,
                        Message = string.Format(
                            CultureInfo.InvariantCulture,
                            "{0} '{1}' has invalid tag value '{2}'!",
                            type,
                            name,
                            tag)
                    });
            }
        }

        [SuppressMessage("Microsoft.Globalization", "CA1303:Do not pass literals as localized parameters")]
        [SuppressMessage("ReSharper", "SuggestBaseTypeForParameter")]
        void VerifyStereotype(XElement element, string type, HashSet<string> validStereotypes)
        {
            if (element == null)
                throw new ArgumentNullException("element");
            if (string.IsNullOrEmpty(type))
                throw new ArgumentNullException("type");
            if (validStereotypes == null)
                throw new ArgumentNullException("validStereotypes");

            if (LogMessageHandler == null)
            {
                return;
            }
            string package, name = GetFullName(element, out package);
            if (string.IsNullOrEmpty(name))
            {
                return;
            }
            var invalidTags =
                element
                .Elements(_uml13 + "ModelElement.stereotype")
                .Descendants(_uml13 + "Stereotype")
                .Where(tv =>
                   tv.Attribute("name") != null
                   &&
                   !validStereotypes.Contains(tv.Attribute("name").Value))
                 .Select(tv => tv.Attribute("name").Value);
            foreach (var tag in invalidTags)
            {
                LogMessageHandler(
                    null,
                    new LogEventArgs
                    {
                        EventType = TraceEventType.Error,
                        Package = package,
                        Message = string.Format(
                            CultureInfo.InvariantCulture,
                            "{0} '{1}' has invalid stereotype '{2}'!",
                            type,
                            name,
                            tag)
                    });
            }
        }

        [SuppressMessage("Microsoft.Globalization", "CA1303:Do not pass literals as localized parameters")]
        void VerifyGeneralization(XContainer model)
        {
            if (model == null)
                throw new ArgumentNullException("model");

            if (LogMessageHandler == null)
            {
                return;
            }
            foreach (var generalization in
                model
                .Descendants(_uml13 + "Generalization")
                .Where(g =>
                   g.Attribute("subtype") != null
                   &&
                   g.Attribute("supertype") != null
                   &&
                   string.Equals(g.Attribute("subtype").Value, g.Attribute("supertype").Value, StringComparison.OrdinalIgnoreCase)))
            {
                string child = generalization
                    .Descendants(_uml13 + "TaggedValue")
                    .Where(t =>
                       t.Attribute("tag") != null
                       &&
                       t.Attribute("tag").Value == "ea_sourceName"
                       &&
                       t.Attribute("value") != null)
                    .Select(t => t.Attribute("value").Value)
                    .FirstOrDefault();
                string parent = generalization
                    .Descendants(_uml13 + "TaggedValue")
                    .Where(t =>
                       t.Attribute("tag") != null
                       &&
                       t.Attribute("tag").Value == "ea_targetName"
                       &&
                       t.Attribute("value") != null)
                    .Select(t => t.Attribute("value").Value)
                    .FirstOrDefault();
                string package = null;
                XElement @namespace;
                if (generalization.Parent != null
                    &&
                    (@namespace = generalization.Parent.Parent) != null
                    &&
                    @namespace.Name == _uml13 + "Package"
                    &&
                    @namespace.Attribute("name") != null)
                {
                    package = @namespace.Attribute("name").Value;
                }
                LogMessageHandler(
                    null,
                    new LogEventArgs
                    {
                        EventType = TraceEventType.Error,
                        Package = package,
                        Message = string.Format(
                            CultureInfo.InvariantCulture,
                            "{0} -> {1} is a reflexive generalization and should be avoided!",
                            child,
                            parent)
                    });
            }
        }

        void UpdateAssociations(XContainer content, XElement association, ref long lastId)
        {
            if (association == null)
                throw new ArgumentNullException("association");

            UpdateAssociationName(association);
            VerifyTag(association, "Association", _associationTags);
            VerifyNotes(content, association, "Association");
            var tagged =
                association
                .Element(_uml13 + "ModelElement.taggedValue");
            if (tagged == null)
            {
                tagged = new XElement("ModelElement.taggedValue");
                association.Add(tagged);
            }
            var styleex =
                tagged
                .Elements(_uml13 + "TaggedValue")
                .Where(e => e.Attributes().Any(a => a.Name == "tag" && a.Value == "styleex"))
                .Select(e => e.Attributes().Single(a => a.Name == "value"))
                .FirstOrDefault();
            if (styleex == null)
            {
                styleex = new XAttribute("value", "");
                tagged.Add(
                    new XElement(
                        _uml13 + "TaggedValue",
                        new XAttribute("tag", "styleex"),
                        styleex));
            }
            if (styleex.Value.IndexOf("alias=", StringComparison.OrdinalIgnoreCase) == -1)
            {
                lastId++;
                styleex.Value += string.Format(CultureInfo.InvariantCulture, "alias={0};", lastId);
            }
            // ReSharper disable once PossibleNullReferenceException
            foreach (var associationEnd in
                association
                .Element(_uml13 + "Association.connection")
                .Elements(_uml13 + "AssociationEnd"))
            {
                UpdateAssociationEnd(associationEnd, ref lastId);
            }
        }

        void UpdateAttributes(XContainer @class, ref long lastId)
        {
            foreach (var attribute in @class.Descendants(_uml13 + "Attribute"))
            {
                VerifyTag(attribute, "Attribute", _attributeTags);
                VerifyNotes(null, attribute, "Attribute");
                VerifyStereotype(attribute, "Attribute", _attributeStereotypes);
                UpdateFeature(attribute, ref lastId);
            }
        }

        void UpdateOperations(XContainer @class, ref long lastId)
        {
            foreach (var operation in @class.Descendants(_uml13 + "Operation"))
            {
                VerifyTag(operation, "Operation", _operationTags);
                VerifyNotes(null, operation, "Operation");
                VerifyStereotype(operation, "Operation", _operationStereotypes);
                UpdateFeature(operation, ref lastId);
            }
        }

        void UpdateClasses(XElement @class, ref long lastId)
        {
            if (@class == null)
                throw new ArgumentNullException("class");

            VerifyTag(@class, "Class", _classTags);
            VerifyNotes(null, @class, "Class");
            VerifyStereotype(@class, "Operation", _classStereotypes);
            var tagged =
                @class
                .Element(_uml13 + "ModelElement.taggedValue");
            if (tagged == null)
            {
                tagged = new XElement("ModelElement.taggedValue");
                @class.Add(tagged);
            }
            var alias =
                tagged
                .Elements(_uml13 + "TaggedValue")
                .Where(e => e.Attributes().Any(a => a.Name == "tag" && a.Value == "alias"))
                .Select(e => e.Attributes().Single(a => a.Name == "value"))
                .FirstOrDefault();
            if (alias == null)
            {
                alias = new XAttribute("value", "");
                tagged.Add(
                    new XElement(
                        _uml13 + "TaggedValue",
                        new XAttribute("tag", "alias"),
                        alias));
            }
            // ReSharper disable once InvertIf
            if (string.IsNullOrWhiteSpace(alias.Value))
            {
                lastId++;
                alias.Value = string.Format(CultureInfo.InvariantCulture, "{0}", lastId);
            }
            UpdateAttributes(@class, ref lastId);
            UpdateOperations(@class, ref lastId);
        }

        [MethodImpl(MethodImplOptions.Synchronized)]
        public void UpdateIds(string inputXmi)
        {
            _ids.Clear();
            var content = XDocument.Load(inputXmi);
            var model =
                content
                .Elements("XMI")
                .Elements("XMI.content")
                .Elements(_uml13 + "Model")
                .First();
            long lastId = FindLastId(model);
            VerifyGeneralization(model);
            var dictionaryItems = FindDictionaryItems(model).ToList();
            VerifyDictionaries(model, dictionaryItems);
            foreach (var @class in
                model
                .Descendants(_uml13 + "Class")
                .Where(e => e.Attributes().All(a => a.Name.LocalName != "name" || a.Value != "EARootClass")))
            {
                UpdateClasses(@class, ref lastId);
            }
            foreach (var association in
                model
                .Descendants(_uml13 + "Association"))
            {
                UpdateAssociations(content, association, ref lastId);
            }
            VerifyMultiplicity(model);
            //VerifyUniquePairs( model );
            foreach (var generalization in
                model
                .Descendants(_uml13 + "Generalization"))
            {
                UpdateGeneralization(generalization, ref lastId);
            }
            var encoding = Encoding.UTF8;
            var settings = new XmlWriterSettings
            {
                Indent = true,
                IndentChars = "\t",
                CloseOutput = false,
                NewLineHandling = NewLineHandling.Entitize,
                Encoding = encoding,
                NewLineOnAttributes = false,
                WriteEndDocumentOnClose = true,
                CheckCharacters = true,
                DoNotEscapeUriAttributes = true,
                ConformanceLevel = ConformanceLevel.Document,
                OmitXmlDeclaration = false,
                NamespaceHandling = NamespaceHandling.OmitDuplicates
            };
            using (var stream = new MemoryStream())
            {
                var xmlWriter = XmlWriter.Create(stream, settings);
                content.WriteTo(xmlWriter);
                xmlWriter.Flush();
                stream.Position = 0;
                using (var file = new FileStream(inputXmi + ".alt", FileMode.Create, FileAccess.Write))
                {
                    var reader = new StreamReader(stream, encoding);
                    var writer = new StreamWriter(file, encoding);
                    string line;
                    while ((line = reader.ReadLine()) != null)
                    {
                        line = line
                            .Replace(" />", "/>")
                            .Replace("encoding=\"Windows-1252\"", "encoding=\"windows-1252\"")
                            .Replace("&#x171;", "&#369;")
                            .Replace("&#x150;", "&#336;")
                            .Replace("&#x151;", "&#337;");
                        writer.WriteLine(line);
                    }
                    writer.Flush();
                }
            }
        }
    }
}
