﻿using System;
using System.CodeDom;
using System.CodeDom.Compiler;
using System.Collections.Generic;
using System.IO;
using System.Linq;
using System.Xml.Linq;
using Microsoft.VisualStudio.TextTemplating;
using Sda.Modeling.Provider;
using Sda.Modeling.Provider.Providers.Xmi;

namespace Sda.Modeling.TextTemplating
{
    public sealed class SdaModelDirectiveProcessor : DirectiveProcessor
    {
        static readonly CodeGeneratorOptions _options = new CodeGeneratorOptions
        {
            BlankLinesBetweenMembers = true,
            IndentString = "  ",
            VerbatimOrder = true,
            BracingStyle = "C"
        };

        readonly StringWriter _classCodeWriter = new StringWriter();
        readonly StringWriter _initializationCodeWriter = new StringWriter();
        string[] _files = new string[0];
        CodeDomProvider _languageProvider;

        public override void Initialize(ITextTemplatingEngineHost host)
        {
            base.Initialize(host);
            TextTemplatingEngineHost textTemplatingEngineHost = host as TextTemplatingEngineHost;
            if (textTemplatingEngineHost != null
                &&
                textTemplatingEngineHost.SourceFiles != null)
            {
                _files = textTemplatingEngineHost.SourceFiles.ToArray();
            }
        }

        public override void StartProcessingRun(CodeDomProvider languageProvider, string templateContents, CompilerErrorCollection errors)
        {
            base.StartProcessingRun(languageProvider, templateContents, errors);
            _languageProvider = languageProvider;
        }

        public override void FinishProcessingRun()
        {
            _languageProvider = null;
        }

        public override string GetClassCodeForProcessingRun()
        {
            return _classCodeWriter.ToString();
        }

        public override string[] GetImportsForProcessingRun()
        {
            return new[]
            {
                "System",
                "System.Xml",
                "System.Xml.Linq",
                typeof( IModel ).Namespace,
                typeof( IModelProvider ).Namespace
            };
        }

        public override string GetPostInitializationCodeForProcessingRun()
        {
            return _initializationCodeWriter.ToString();
        }

        public override string GetPreInitializationCodeForProcessingRun()
        {
            return "";
        }

        public override string[] GetReferencesForProcessingRun()
        {
            return new[]
            {
                "System.Xml",
                "System.Xml.Linq",
                "System.Core",
                typeof( IModel ).Assembly.Location,
                typeof( IModelProvider ).Assembly.Location
            };
        }

        public override bool IsDirectiveSupported(string directiveName)
        {
            return string.Equals(directiveName, @"sdamodel", StringComparison.OrdinalIgnoreCase);
        }

        public override void ProcessDirective(string directiveName, IDictionary<string, string> arguments)
        {
            string xmiArgument;
            if (_files.Length == 0
                &&
                arguments.TryGetValue("Xmi", out xmiArgument))
            {
                _files = xmiArgument.Split(new[] { ";" }, StringSplitOptions.RemoveEmptyEntries);
            }
            string value;
            var project = arguments.TryGetValue("Project", out value) ? value : "";
            var modelField = new CodeMemberField(typeof(IModel), "model")
            {
                Attributes = MemberAttributes.Private,
            };
            _languageProvider.GenerateCodeFromMember(modelField, _classCodeWriter, _options);
            var modelProperty = new CodeMemberProperty
            {
                Name = "Model",
                Type = new CodeTypeReference(typeof(IModel)),
                Attributes = MemberAttributes.Public,
                HasGet = true,
                HasSet = false,
            };
            modelProperty.GetStatements.Add(
                new CodeMethodReturnStatement(
                    new CodeFieldReferenceExpression(
                        new CodeThisReferenceExpression(),
                        "model")));
            var projectNameProperty = new CodeMemberProperty
            {
                Name = "ProjectName",
                Type = new CodeTypeReference(typeof(string)),
                Attributes = MemberAttributes.Public,
                HasGet = true,
                HasSet = false,
            };
            projectNameProperty.GetStatements.Add(
                new CodeMethodReturnStatement(
                    new CodePrimitiveExpression(project)));
            _languageProvider.GenerateCodeFromMember(projectNameProperty, _classCodeWriter, _options);
            _languageProvider.GenerateCodeFromMember(modelProperty, _classCodeWriter, _options);
            var docs = _files
                .Select(f =>
                   new CodeMethodInvokeExpression(
                       new CodeTypeReferenceExpression(typeof(XDocument)),
                       "Load",
                       new CodePrimitiveExpression(f)))
                .OfType<CodeExpression>()
                .ToArray();
            var assignment = new CodeAssignStatement(
                new CodeFieldReferenceExpression(
                    new CodeThisReferenceExpression(), "model"),
                new CodeCastExpression(
                    new CodeTypeReference(typeof(IModel)),
                    new CodeMethodInvokeExpression(
                        new CodeObjectCreateExpression(
                            new CodeTypeReference(
                                typeof(XmiModelProvider)),
                                docs
                            ),
                        "Model",
                        new CodePrimitiveExpression(project))));
            _languageProvider.GenerateCodeFromStatement(assignment, _initializationCodeWriter, _options);
        }
    }
}
